from typing import List, Mapping, Generator, Tuple, Optional, Any
from trialbot.data.ns_vocabulary import NSVocabulary
from ._translation_base import _TranslationBase, NullableTensor
from .field import Field
import torch
from collections import defaultdict

class Translator(_TranslationBase):
    """
    A Translator has the full knowledge of the accompanied dataset.
    It has to provide APIs to convert a dataset example into a tensor based storage.
    """
    def batch_tensor(self, tensors: List[Mapping[str, NullableTensor]]) -> Mapping[str, torch.Tensor]:
        raise NotImplementedError


class FieldAwareTranslator(Translator):
    def __init__(self, field_list: List[Field], vocab_fields: Optional[List[Field]] = None):
        """ Initialize the translator with some fields.
        field_list: default fields
        vocab_fields: only used to build the vocabulary indices, default to field_list if not applicable
        """
        super().__init__()
        self.fields = field_list
        self.vocab_fields = vocab_fields or field_list

    def index_with_vocab(self, vocab: NSVocabulary):
        super().index_with_vocab(vocab)
        for field in self.fields:
            field.index_with_vocab(vocab)

    def generate_namespace_tokens(self, example) -> Generator[Tuple[str, str], None, None]:
        for field in self.vocab_fields:
            yield from field.generate_namespace_tokens(example)

    def to_tensor(self, example) -> Mapping[str, NullableTensor]:
        instance_fields = {}
        for field in self.fields:
            instance_fields.update(field.to_tensor(example))
        return instance_fields

    def batch_tensor(self, tensors: List[Mapping[str, NullableTensor]]) -> Mapping[str, torch.Tensor]:
        tensors = list(filter(lambda x: all(v is not None for v in x.values()), tensors))
        batch_dict = self.list_of_dict_to_dict_of_list(tensors)
        output = {}
        for field in self.fields:
            output.update(field.batch_tensor_by_key(batch_dict))
        return output

    @staticmethod
    def list_of_dict_to_dict_of_list(ld: List[Mapping[str, Any]]) -> Mapping[str, List[Any]]:
        list_by_keys = defaultdict(list)
        for d in ld:
            for k, v in d.items():
                list_by_keys[k].append(v)
        return list_by_keys





import numpy
from ..iterator import Iterator
import logging


class RandomIterator(Iterator):
    """Dataset iterator that serially reads the examples.

    This is a simple implementation of :class:`~chainer.dataset.Iterator`
    that just visits each example in either the order of indexes or a shuffled
    order.

    To avoid unintentional performance degradation, the ``shuffle`` option is
    set to ``True`` by default. For validation, it is better to set it to
    ``False`` when the underlying dataset supports fast slicing. If the
    order of examples has an important meaning and the updater depends on the
    original order, this option should be set to ``False``.

    Args:
        dataset: Dataset to iterate.
        batch_size (int): Number of examples within each batch.
        repeat (bool): If ``True``, it infinitely loops over the dataset.
            Otherwise, it stops iteration at the end of the first epoch.
        shuffle (bool): If ``True``, the order of examples is shuffled at the
            beginning of each epoch. Otherwise, examples are extracted in the
            order of indexes. If ``None`` and no ``order_sampler`` is given,
            the behavior is the same as the case with ``shuffle=True``.
    """

    def __init__(self, dataset_len: int, batch_size: int, repeat: bool = True, shuffle: bool = True):
        super(RandomIterator, self).__init__()
        self.dataset_len = dataset_len
        self.batch_size = batch_size
        self.repeat: bool = repeat
        self.shuffle: bool = shuffle
        self._order = None if not self.shuffle else numpy.random.permutation(numpy.arange(self.dataset_len))
        self.current_position = 0
        self.logger = logging.getLogger(__name__)
        self.epoch = 0
        self.is_end_of_epoch = False

    def __next__(self):
        if self._epoch_size == 0 or self.batch_size == 0:
            self.logger.warning("Given dataset is empty. Nothing to yield but only raising StopIteration")
            raise StopIteration

        if not self.repeat and self.epoch > 0:
            raise StopIteration

        i = self.current_position
        i_end = i + self.batch_size
        N = self._epoch_size

        if self._order is None:
            batch = list(range(self.dataset_len))[i:i_end]
        else:
            batch = [index for index in self._order[i:i_end]]

        if i_end >= N:
            if self.repeat:
                rest = i_end - N
                if self._order is not None:
                    new_order = numpy.random.permutation(numpy.arange(self.dataset_len))
                    self._order = new_order

                if rest > 0:
                    if self._order is None:
                        batch.extend(list(range(rest)))
                    else:
                        batch.extend([index for index in self._order[:rest]])
                self.current_position = rest
            else:
                self.current_position = 0

            self.epoch += 1
            self.is_end_of_epoch = True
        else:
            self.is_end_of_epoch = False
            self.current_position = i_end

        return batch

    next = __next__

    def reset(self, skip=0):
        self.current_position = skip
        self.epoch = 0
        self.is_end_of_epoch = False

        if self.shuffle:
            self._order = numpy.random.permutation(numpy.arange(self.dataset_len))

    @property
    def _epoch_size(self):
        if self._order is None:
            return self.dataset_len
        else:
            return len(self._order)




import torch
import random
import numpy as np

def fix_seed(seed: int):
    """Sets random seed everywhere."""
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
    random.seed(seed)
    np.random.seed(seed)


from .trial_registry import Registry
from .trial_bot import TrialBot, Events, State
from .updater import Updater
from .updaters.training_updater import TrainingUpdater
from .updaters.testing_updater import TestingUpdater

from . import extensions as bot_extensions



from typing import List, Mapping, Callable
from trialbot.training import Registry
from itertools import product


def expand_hparams(grid_conf: Mapping[str, list]):
    keys, value_lists = zip(*grid_conf.items())
    for v in product(*value_lists):
        params = dict(zip(keys, v))
        yield params

def update_registry_params(name_prefix, i, params, base_params_fn: Callable):
    name = f"{name_prefix}{i}"

    def hparams_wrapper_fn():
        p = base_params_fn()
        for k, v in params.items():
            setattr(p, k, v)
        return p

    Registry._hparamsets[name] = hparams_wrapper_fn
    return name

def import_grid_search_parameters(grid_conf: Mapping[str, list], base_param_fn: Callable, name_prefix: str = None):
    param_sets = list(expand_hparams(grid_conf))
    import re
    if name_prefix is None:
        name_prefix = re.sub('[^a-zA-Z_0-9]', '', base_param_fn.__name__) + '_'
    names = [update_registry_params(name_prefix, i, params, base_param_fn) for i, params in enumerate(param_sets)]
    return names

if __name__ == '__main__':
    grid_conf = {"batch_size": [64, 128, 256], "hidden_size": [300, 600]}
    grid_conf = {
                    "num_inner_loops": [1, 2, 3, 5, 10],
                    "batch_sz": [20, 40, 80],
                    "support_batch_sz": [100, 200, 400]
                }
    from trialbot.training.hparamset import HyperParamSet
    names = import_grid_search_parameters(grid_conf, lambda: HyperParamSet())
    print(f"names={names}")
    print(Registry._hparamsets)
    p = Registry.get_hparamset(names[-1])
    print(type(p))

    for name in names:
        print('--------' * 10)
        print("hparamset: " + name)
        print(Registry.get_hparamset(name))




import datetime
import torch
import os.path

def ext_write_info(bot, msg):
    bot.logger.info(msg)

def current_epoch_logger(bot):
    bot.logger.info(f"Current Epoch {bot.state.epoch}: {datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}")

def time_logger(bot):
    bot.logger.info(f"Current Time {datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}")

def every_epoch_model_saver(bot, interval=1):
    if bot.state.epoch % interval == 0:
        savedir, model = bot.savepath, bot.model
        filename = os.path.join(savedir, f"model_state_{bot.state.epoch}.th")
        torch.save(model.state_dict(), filename)
        bot.logger.info(f"model saved to {filename}")

def loss_reporter(bot, interval=4):
    if bot.state.iteration % interval != 0:
        return

    import torch
    import numpy
    output = bot.state.output
    if output is None:
        return

    loss = output.get('loss')
    if isinstance(loss, torch.Tensor):
        loss = loss.item()
    elif isinstance(loss, numpy.ndarray):
        loss = loss.item()
    else:
        return

    bot.logger.info(f"Epoch: {bot.state.epoch}, Iteration: {bot.state.iteration}, Loss: {loss:.4f}")

def legacy_testing_output(bot):
    import json
    output = bot.state.output
    if output is None:
        return
    model = bot.model
    output = model.decode(output)
    print(json.dumps(output['predicted_tokens']))



from typing import List, Mapping, Union, Callable, Any

import random

import logging
from collections import defaultdict

from .bucket_iterator import BucketIterator

class ClusterIterator(BucketIterator):
    def __init__(self,
                 dataset,
                 batch_size,
                 cluster_key: Union[str, Callable[[Any], str]],
                 repeat=True,
                 shuffle=True,
                 ):
        super().__init__(dataset, batch_size, cluster_key, repeat=repeat, shuffle=shuffle)
        self.logger = logging.getLogger(self.__class__.__name__)
        self._clusters = None
        if isinstance(cluster_key, str):
            self._key_func = lambda x: x[cluster_key]

    def _get_batches(self):
        if self._clusters is None:
            self._clusters = defaultdict(list)
            for i, x in enumerate(self.dataset):
                key = self._key_func(x)
                self._clusters[key].append(i)

        groups = list(self._clusters.values())
        if self._shuffle:
            random.shuffle(groups)

        batches = []
        for g in groups:
            items = [i for i in g]
            random.shuffle(items)
            for i in range(0, len(items), self.batch_size):
                batch = items[i:i+self.batch_size]
                batches.append(batch)
        return batches



from typing import Optional, List, Dict, Union, Any
import torch.nn
from .trial_bot import TrialBot
from trialbot.data.iterator import Iterator

class Updater:
    def __init__(self,
                 models: Union[List[torch.nn.Module], torch.nn.Module],
                 iterators: Union[List[Iterator], Iterator],
                 optims,
                 device: int = -1):
        self._models = models if isinstance(models, list) else [models]
        self._device = device
        self._iterators: List[Iterator] = iterators if isinstance(iterators, list) else [iterators]
        self._optims = optims if isinstance(optims, list) else [optims]

        self._epoch_ended = False

    def start_epoch(self):
        self._epoch_ended = False

    def __call__(self):
        """
        If called by __call__, Updater will returns a iterator.
        Therefore supporting the following paradigm for training or testing loops.

        ```python
        for output in updater():
            # do stuff with output
        ```

        The effect is the same as iter(updater).

        If you want only the next example, use next(updater) instead.

        :return:
        """
        yield from self

    def __iter__(self):
        """Return self. Support iteration for only once."""
        return self

    def __next__(self):
        if self._epoch_ended:
            raise StopIteration

        return self.update_epoch()

    def update_epoch(self):
        """
        Keep updating until the epoch ends.
        When the epoch has ended, the method should call stop_epoch.
        :return:
        """
        raise NotImplementedError

    def stop_epoch(self):
        self._epoch_ended = True

    @classmethod
    def from_bot(cls, bot: TrialBot) -> 'Updater':
        raise NotImplementedError




from typing import Union, Optional
import os.path

def find_root(root_indicator_file: str = ".ROOT",
              startpoint: Optional[str] = None):
    if startpoint is None:
        startpoint = os.getcwd()
    else:
        startpoint = os.path.normpath(os.path.abspath(startpoint))

    rootpath = check_dir_(root_indicator_file, startpoint)
    if rootpath is None:
        raise ValueError('ROOT directory not found until the root of the drive.')

    # return relative path in case of directory privacy
    return os.path.relpath(rootpath, os.curdir)

def check_dir_(root_filename: str, dirpath: str):
    if root_filename in os.listdir(dirpath):
        return dirpath

    # go upper level
    parent = get_upper_level_dir_(dirpath)
    if parent == dirpath:   # root of fs encountered
        return None

    return check_dir_(root_filename, parent)

def get_upper_level_dir_(dirpath: str) -> str:
    canonical_dirpath = os.path.normpath(os.path.abspath(dirpath))
    parent = os.path.join(canonical_dirpath, os.path.pardir)
    canonical_parent = os.path.normpath(parent)

    return canonical_parent







from typing import List
from trialbot.data.dataset import Dataset

class IndexDataset(Dataset):
    def __init__(self, dataset_store: Dataset, index: List[int]):
        self._store = dataset_store
        self.index = index

    def __len__(self):
        return len(self.index)

    def get_example(self, i: int):
        return self._store[self.index[i]]


from typing import Callable, Optional, Dict, Any
import torch.nn
import os.path
import json
from datetime import datetime

from trialbot.data.dataset import Dataset
from trialbot.data.ns_vocabulary import NSVocabulary
from tqdm import tqdm
from enum import Enum

from trialbot.data.translator import Translator

from .opt_parser import get_trial_bot_common_opt_parser
from .trial_registry import Registry
from .event_engine import Engine
from . import extensions as ext_mod
import logging
logging.basicConfig()

class Events(Enum):
    """Predefined events for TrialBot training."""
    EPOCH_STARTED = "epoch_started"
    EPOCH_COMPLETED = "epoch_completed"
    STARTED = "started"
    COMPLETED = "completed"
    ITERATION_STARTED = "iteration_started"
    ITERATION_COMPLETED = "iteration_completed"
    EXCEPTION_RAISED = "exception_raised"

class State(object):
    """An object that is used to pass internal and user-defined state between event handlers."""
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

class TrialBot:
    """
    A TrialBot is the direct manager for an experiment, which is responsible for only two requirements,
    Run and Serve, corresponding to run the experiment or to get loaded by some server.

    A TrialBot maintains several Components, and schedules the experiment procedure.
    Components are as follows so far,
    - args: arguments parsed from commandline
    - hparamset: the designated hyperparameters from the running shell
    - dataset: specifically train_set, dev_set, test_set
    - translator: how the dataset should be interpreted.
    - savepath: where the data will be saved if given, or set by default.
    - vocab: the vocabulary with several namespaces
    - models: the list in which all the models for the single experiments are included.

    An experiment is empirically divided into two phase, Training and Testing.
    They share the same program entry _run_ and basically consist of the similar routines,
    1. Do some preprocessing job if any (e.g. save vocab for reusing, makedirs for savepath, etc.)
    2. Initialize the updater for running the experiment.
        1) For a standard updater, the classmethod of the updater already implements the initialization
        of the iterators and optimizers.
        2) For a custom updater, the class initialization could be anywhere given the bot with its components.
        Then the custom updater must be injected into the bot with property assignment. like

        ```python
            bot.updater = Updater(bot)
        ```
    3. Run the looping until the training ends. This starts an event-driven engine and fires events
       at some time. All the DIY requirements could be implemented by extensions, which are registered
       to the engine for some specific events and get run by the event engine when the very event is
       fired.
       By default, the loop contains some epochs and some iterations of each epoch.
       The loop relies on the updater, and start a new epoch only after the updater raise an StopIteration exception.

    """
    def __init__(self,
                 args = None,
                 trial_name="default_savedir",
                 get_model_func: Optional[Callable[[Any, NSVocabulary], torch.nn.Module]] = None,
                 clean_engine: bool = False,
                 ):
        if args is None:
            parser = TrialBot.get_default_parser()
            args = parser.parse_args()

        self.args = args
        self.name = trial_name

        # self.translator = None
        # self.vocab = None
        # self.train_set = None
        # self.dev_set = None
        # self.test_set = None
        self.hparams = None
        self.savepath = "."
        self.logger = logging.getLogger(__name__)

        self.updater = None

        self.state = State(epoch=0, iteration=0, output=None)

        self.get_model = get_model_func

        self._engine = self._make_engine(clean=clean_engine)
        self._init_components()

    def _make_engine(self, clean: bool = False):
        engine = Engine()
        engine.register_events(*Events)
        # events with greater priorities will get processed earlier.
        if not clean:
            engine.add_event_handler(Events.EPOCH_STARTED, ext_mod.ext_write_info, 100, msg="Epoch started")
            engine.add_event_handler(Events.EPOCH_STARTED, ext_mod.ext_write_info, 105, msg=("====" * 20))
            engine.add_event_handler(Events.EPOCH_STARTED, ext_mod.current_epoch_logger, 99)
            engine.add_event_handler(Events.STARTED, ext_mod.ext_write_info, 100, msg="TrailBot started")
            engine.add_event_handler(Events.STARTED, ext_mod.time_logger, 99)
            engine.add_event_handler(Events.COMPLETED, ext_mod.time_logger, 101)
            engine.add_event_handler(Events.COMPLETED, ext_mod.ext_write_info, 100, msg="TrailBot completed.")
            engine.add_event_handler(Events.ITERATION_COMPLETED, ext_mod.loss_reporter, 100)
        return engine

    @property
    def model(self):
        return self.models[0]

    @staticmethod
    def get_default_parser():
        parser = get_trial_bot_common_opt_parser()
        parser.add_argument('models', nargs='*', help='pretrained models for the same setting')
        parser.add_argument('--dry-run', action="store_true")
        parser.add_argument('--skip', type=int, help='skip NUM examples for the first iteration, intended for debug use.')
        parser.add_argument('--vocab-dump', help="the file path to save and load the vocab obj")
        return parser

    def _init_components(self):
        """
        Start a trial directly.
        """
        args = self.args

        # logging args
        if args.quiet:
            self.logger.setLevel(logging.WARNING)
        elif args.debug:
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.INFO)

        hparams = Registry.get_hparamset(args.hparamset)
        self.hparams = hparams
        savepath = args.snapshot_dir if args.snapshot_dir else (os.path.join(
            hparams.SNAPSHOT_PATH,
            args.dataset,
            self.name,
            datetime.now().strftime('%Y%m%d-%H%M%S') + ('-' + args.memo if args.memo else '')
        ))
        self.savepath = savepath

        self._init_dataset()
        self._init_translator()

        self.vocab = self._init_vocab(self.train_set, self.translator)

        self.translator.index_with_vocab(self.vocab)
        self.models = self._init_models(hparams, self.vocab)

    def _init_dataset(self):
        args = self.args
        self.datasets = Registry.get_dataset(args.dataset)
        return self.datasets

    @property
    def train_set(self):
        return self._get_classical_dataset_split(0)

    @property
    def dev_set(self):
        return self._get_classical_dataset_split(1)

    @property
    def test_set(self):
        return self._get_classical_dataset_split(2)

    def _get_classical_dataset_split(self, i):
        assert len(self.datasets) >= 3 and all(isinstance(x, Dataset) for x in self.datasets[:3])
        return self.datasets[i]

    def _init_translator(self):
        args, p = self.args, self.hparams
        translator_kwargs = getattr(p, 'TRANSLATOR_KWARGS', dict())
        translator = Registry.get_translator(args.translator, **translator_kwargs)
        self.translator = translator
        return translator

    def _init_vocab(self,
                    dataset: Dataset,
                    translator: Translator):
        args, logger = self.args, self.logger
        hparams = self.hparams

        if args.vocab_dump and os.path.exists(args.vocab_dump):
            logger.info(f"read vocab from file {args.vocab_dump}")
            vocab = NSVocabulary.from_files(args.vocab_dump)

        else:
            logger.info("initialize vocab from training data")
            # count for a counter
            logger.debug("start initializing vocab")
            counter: Dict[str, Dict[str, int]] = dict()
            for example in tqdm(iter(dataset)):
                for namespace, w in translator.generate_namespace_tokens(example):
                    if namespace not in counter:
                        counter[namespace] = dict()

                    if w not in counter[namespace]:
                        counter[namespace][w] = 0

                    counter[namespace][w] += 1

            vocab = NSVocabulary(counter, **hparams.NS_VOCAB_KWARGS)

        if args.vocab_dump:
            os.makedirs(args.vocab_dump, exist_ok=True)
            vocab.save_to_files(args.vocab_dump)
        logger.info(str(vocab))

        return vocab

    def _init_models(self, hparams, vocab):
        args = self.args
        models = self.get_model(hparams, vocab)
        if not isinstance(models, list) and not isinstance(models, tuple):
            models = [models]

        if args.models:
            for model, model_path in zip(models, args.models):
                model.load_state_dict(torch.load(model_path))

        if args.device >= 0:
            models = [model.cuda(args.device) for model in models]
        return models

    def run(self, training_epoch: int = 0):
        if training_epoch == 0:
            training_epoch = self.hparams.TRAINING_LIMIT

        if self.args.test:
            # testing procedure
            # 1. init updater; 2. run the testing engine
            if self.updater is None:
                from .updaters.testing_updater import TestingUpdater
                self.updater = TestingUpdater.from_bot(self)
            self._testing_engine_loop(self.updater)

        else:
            # training procedure
            # 1. save vocab; 2. init updater; 3. start main engine
            savepath = self.savepath
            if not os.path.exists(savepath):
                os.makedirs(savepath, mode=0o755)
            vocab_path = os.path.join(savepath, 'vocab')
            if not os.path.exists(vocab_path):
                os.makedirs(vocab_path, mode=0o755)
                self.vocab.save_to_files(vocab_path)

            if self.updater is None:
                from .updaters.training_updater import TrainingUpdater
                self.updater = TrainingUpdater.from_bot(self)
            self._training_engine_loop(self.updater, training_epoch)

    def _training_engine_loop(self, updater, max_epoch):
        engine = self._engine
        engine.fire_event(Events.STARTED, bot=self)
        while self.state.epoch < max_epoch:
            self.state.epoch += 1
            engine.fire_event(Events.EPOCH_STARTED, bot=self)
            updater.start_epoch()
            while True:
                self.state.iteration += 1
                engine.fire_event(Events.ITERATION_STARTED, bot=self)
                try:
                    self.state.output = next(updater)
                except StopIteration:
                    self.state.iteration -= 1
                    self.state.output = None
                    break
                finally:
                    engine.fire_event(Events.ITERATION_COMPLETED, bot=self)

            engine.fire_event(Events.EPOCH_COMPLETED, bot=self)
        engine.fire_event(Events.COMPLETED, bot=self)

    def _testing_engine_loop(self, updater):
        engine = self._engine
        with torch.no_grad():
            engine.fire_event(Events.STARTED, bot=self)
            engine.fire_event(Events.EPOCH_STARTED, bot=self)
            updater.start_epoch()
            while True:
                self.state.iteration += 1
                engine.fire_event(Events.ITERATION_STARTED, bot=self)
                try:
                    self.state.output = next(updater)
                except StopIteration:
                    self.state.iteration -= 1
                    self.state.output = None
                    break
                finally:
                    engine.fire_event(Events.ITERATION_COMPLETED, bot=self)

            engine.fire_event(Events.EPOCH_COMPLETED, bot=self)
            engine.fire_event(Events.COMPLETED, bot=self)

    def attach_extension(self,
                         event_name: str = Events.ITERATION_COMPLETED,
                         priority: int = 100,):
        """Used as a decorator only. To add extension directly, use add_event_handler instead."""
        def decorator(handler):
            self._engine.add_event_handler(event_name, handler, priority)
            return handler
        return decorator

    def add_event_handler(self, *args, **kwargs):
        self._engine.add_event_handler(*args, **kwargs)




from typing import List, Mapping, Generator, Tuple, Optional, Any, Literal
from ..field import Field, NullableTensor
import torch
from trialbot.data import START_SYMBOL, END_SYMBOL, PADDING_TOKEN, NSVocabulary
from itertools import product
from torch.nn.utils.rnn import pad_sequence

class SeqField(Field):
    def get_sent(self, example):
        return example.get(self.source_key)

    def generate_namespace_tokens(self, example) -> Generator[Tuple[str, str], None, None]:
        if self.add_start_end_toks:
            yield from product([self.ns], [START_SYMBOL, END_SYMBOL])

        seq_raw = self.get_sent(example)
        if seq_raw is not None:
            yield from product([self.ns], (x.lower() if self.lower_case else x for x in self.split(seq_raw)))

    def to_tensor(self, example) -> Mapping[str, NullableTensor]:
        seq_raw = self.get_sent(example)
        if seq_raw is None:
            return {self.renamed_key: torch.tensor([self.padding])}

        seq_toks = self.split(seq_raw)
        if len(seq_toks) == 0:
            return {self.renamed_key: torch.tensor([self.padding])}

        if self.max_seq_len > 0:
            seq_toks = seq_toks[:self.max_seq_len]

        if self.lower_case:
            seq_toks = [x.lower() if self.lower_case else x for x in seq_toks]

        if self.add_start_end_toks:
            seq_toks = [START_SYMBOL] + seq_toks + [END_SYMBOL]

        seq_toks = [self.vocab.get_token_index(tok, self.ns) for tok in seq_toks]
        seq_tensor = torch.tensor(seq_toks)
        return {self.renamed_key: seq_tensor}

    def batch_tensor_by_key(self,
                            tensors_by_keys: Mapping[str, List[NullableTensor]]
                            ) -> Mapping[str, torch.Tensor]:
        tensor_list = tensors_by_keys.get(self.renamed_key)
        if tensor_list is None or len(tensor_list) == 0:
            raise KeyError(f'Empty field key {self.renamed_key} confronted. Failed to build the instance tensors')

        batch_tensor = pad_sequence(tensor_list, batch_first=True, padding_value=self.padding)
        return {self.renamed_key: batch_tensor}

    def __init__(self, source_key: str,
                 renamed_key: str = None,
                 namespace: str = None,
                 split_fn=None,
                 add_start_end_toks: bool = True,
                 padding_id: int = 0,
                 max_seq_len: int = 0,
                 use_lower_case: bool = True,
                 ):
        super().__init__()
        self.split = split_fn or (lambda s: s if isinstance(s, list) else s.split())
        self.source_key = source_key
        self.ns = namespace or source_key
        self.renamed_key = renamed_key or source_key
        self.add_start_end_toks = add_start_end_toks
        self.padding = padding_id
        self.max_seq_len = max_seq_len
        self.lower_case = use_lower_case



from trialbot.utils.move_to_device import move_to_device
import logging
from ..trial_bot import TrialBot
from trialbot.data.iterators import RandomIterator
from ..updater import Updater
from trialbot.data.translator import Translator

class TestingUpdater(Updater):
    def __init__(self, dataset, translator: Translator, models, iterators, optims, device=-1):
        super().__init__(models, iterators, optims, device)
        self.dataset = dataset
        self.translator: Translator = translator

    def update_epoch(self):
        model, iterator, device = self._models[0], self._iterators[0], self._device
        model.eval()

        indices = next(iterator)
        tensor_list = [self.translator.to_tensor(self.dataset[index]) for index in indices]
        try:
            batch = self.translator.batch_tensor(tensor_list)
        except Exception as e:
            logging.getLogger(self.__class__.__name__).warning(
                f'skipping the preprocessing of the batch which raises an exception ... {str(e)}'
            )
            return None

        if iterator.is_end_of_epoch:
            self.stop_epoch()
        if device >= 0:
            batch = move_to_device(batch, device)
        output = model(**batch)
        return output

    @classmethod
    def from_bot(cls, bot: TrialBot) -> 'TestingUpdater':
        args, model, hparams = bot.args, bot.model, bot.hparams
        device = args.device
        dataset = bot.dev_set if args.dev else bot.test_set
        iterator = RandomIterator(len(dataset), hparams.batch_sz, shuffle=False, repeat=False)
        updater = cls(dataset, bot.translator, model, iterator, None, device)
        return updater




from ..dataset import Dataset

import trialbot.utils.file_reader as reader_utils

class FileDataset(Dataset):
    def __init__(self, filename, lazy=True):
        super(FileDataset, self).__init__()
        self.lazy = lazy

        self.filename = filename

        self._data = None
        if not lazy:
            self._read_data()

    def _read_data(self):
        if self._data is None:
            self._data = list(line.rstrip('\r\n') for line in reader_utils.open_file(self.filename))

    def __len__(self):
        self._read_data()
        return len(self._data)

    def get_example(self, i):
        self._read_data()
        return self._data[i]

class DirectoryDataset(Dataset):
    def __init__(self, dirname):
        super(DirectoryDataset, self).__init__()
        self.dirname = dirname
        self._data = list(line.rstrip('\r\n') for line in reader_utils.open_dir(dirname))

    def __len__(self):
        return len(self._data)

    def get_example(self, i):
        return self._data[i]




from typing import List, Mapping, Optional
import torch
from ._translation_base import _TranslationBase, NullableTensor

class Field(_TranslationBase):
    """
    A field is not actually a translator, this is experimental
    and will later be refactored to inherit from another interface.
    """
    def batch_tensor_by_key(self,
                            tensors_by_keys: Mapping[str, List[NullableTensor]]
                            ) -> Mapping[str, torch.Tensor]:
        """
        A normal translator accept a batch of tensors,
        which is an iterable collection containing the instances from Translator.to_tensor interface.
        A field will handle the tensor list by some predefined key,
        but will accept all the possible tensors batch dict by the keys.
        :param tensors_by_keys:
        :return:
        """
        raise NotImplementedError




"""
This code file is majorly based on pytorch-ignite, whose license is available at
https://github.com/pytorch/ignite/blob/master/LICENSE,
and duplicated below.

BSD 3-Clause License

Copyright (c) 2018, PyTorch team
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import logging
from collections import defaultdict

class Engine(object):
    """Runs a given process_function over each batch of a dataset, emitting events as it goes.

    Args:
        process_function (callable): A function receiving a handle to the engine and the current batch
            in each iteration, and returns data to be stored in the engine's state.

    Example usage:

    .. code-block:: python

        def train_and_store_loss(engine, batch):
            inputs, targets = batch
            optimizer.zero_grad()
            outputs = model(inputs)
            loss = loss_fn(outputs, targets)
            loss.backward()
            optimizer.step()
            return loss.item()

        engine = Engine(train_and_store_loss)
        engine.run(data_loader)

        # Loss value is now stored in `engine.state.output`.

    """
    def __init__(self):
        self._event_handlers = defaultdict(list)
        self._logger = logging.getLogger(__name__ + "." + self.__class__.__name__)
        self._logger.addHandler(logging.NullHandler())
        self._allowed_events = []

    def register_events(self, *event_names):
        """Add events that can be fired.

        Registering an event will let the user fire these events at any point.
        This opens the door to make the :meth:`~ignite.engine.Engine.run` loop even more
        configurable.

        By default, the events from :class:`~ignite.engine.Events` are registerd.

        Args:
            *event_names: An object (ideally a string or int) to define the
                name of the event being supported.

        Example usage:

        .. code-block:: python

            from enum import Enum

            class Custom_Events(Enum):
                FOO_EVENT = "foo_event"
                BAR_EVENT = "bar_event"

            engine = Engine(process_function)
            engine.register_events(*Custom_Events)

        """
        for name in event_names:
            self._allowed_events.append(name)

    def add_event_handler(self, event_name, handler, priority, *args, **kwargs):
        """Add an event handler to be executed when the specified event is fired.

        Args:
            event_name: An event to attach the handler to. Valid events are from :class:`~ignite.engine.Events`
                or any `event_name` added by :meth:`~ignite.engine.Engine.register_events`.
            handler (callable): the callable event handler that should be invoked
            *args: optional args to be passed to `handler`.
            **kwargs: optional keyword args to be passed to `handler`.

        Notes:
              The handler function's first argument will be `self`, the :class:`~ignite.engine.Engine` object it
              was bound to.

              Note that other arguments can be passed to the handler in addition to the `*args` and  `**kwargs`
              passed here, for example during :attr:`~ignite.engine.Events.EXCEPTION_RAISED`.

        Example usage:

        .. code-block:: python

            engine = Engine(process_function)

            def print_epoch(engine):
                print("Epoch: {}".format(engine.state.epoch))

            engine.add_event_handler(Events.EPOCH_COMPLETED, print_epoch)

        """
        if event_name not in self._allowed_events:
            self._logger.error("attempt to add event handler to an invalid event %s.", event_name)
            raise ValueError("Event {} is not a valid event for this Engine.".format(event_name))

        self._event_handlers[event_name].append((handler, priority, args, kwargs))
        self._logger.debug("added handler for event %s.", event_name)

    def has_event_handler(self, handler, event_name=None):
        """Check if the specified event has the specified handler.

        Args:
            handler (callable): the callable event handler.
            event_name: The event the handler attached to. Set this
                to ``None`` to search all events.
        """
        if event_name is not None:
            if event_name not in self._event_handlers:
                return False
            events = [event_name]
        else:
            events = self._event_handlers
        for e in events:
            for h, _, _, _ in self._event_handlers[e]:
                if h == handler:
                    return True
        return False

    def remove_event_handler(self, handler, event_name):
        """Remove event handler `handler` from registered handlers of the engine

        Args:
            handler (callable): the callable event handler that should be removed
            event_name: The event the handler attached to.

        """
        if event_name not in self._event_handlers:
            raise ValueError("Input event name '{}' does not exist".format(event_name))

        new_event_handlers = [(h, p, args, kwargs) for h, p, args, kwargs in self._event_handlers[event_name]
                              if h != handler]
        if len(new_event_handlers) == len(self._event_handlers[event_name]):
            raise ValueError("Input handler '{}' is not found among registered event handlers".format(handler))
        self._event_handlers[event_name] = new_event_handlers

    def on(self, event_name, *args, **kwargs):
        """Decorator shortcut for add_event_handler.

        Args:
            event_name: An event to attach the handler to. Valid events are from :class:`~ignite.engine.Events` or
                any `event_name` added by :meth:`~ignite.engine.Engine.register_events`.
            *args: optional args to be passed to `handler`.
            **kwargs: optional keyword args to be passed to `handler`.

        """
        def decorator(f):
            self.add_event_handler(event_name, f, *args, **kwargs)
            return f
        return decorator

    def fire_event(self, event_name, *event_args, **event_kwargs):
        """Execute all the handlers associated with given event.

        This method executes all handlers associated with the event
        `event_name`. Optional positional and keyword arguments can be used to
        pass arguments to **all** handlers added with this event. These
        aguments updates arguments passed using :meth:`~ignite.engine.Engine.add_event_handler`.

        Args:
            event_name: event for which the handlers should be executed. Valid
                events are from :class:`~ignite.engine.Events` or any `event_name` added by
                :meth:`~ignite.engine.Engine.register_events`.
            *event_args: optional args to be passed to all handlers.
            **event_kwargs: optional keyword args to be passed to all handlers.

        """
        if event_name in self._allowed_events:
            self._logger.debug("firing handlers for event %s ", event_name)
            for func, _, args, kwargs in sorted(self._event_handlers[event_name], key=lambda x: x[1], reverse=True):
                kwargs.update(event_kwargs)
                func(*(event_args + args), **kwargs)



from typing import Callable, Tuple

from trialbot.data.dataset import Dataset
from trialbot.data.translator import Translator
from .hparamset import HyperParamSet

class Registry:
    """
    A static class contains references to all registered
    datasets, translators, and hparamsets by the user script.
    """
    _datasets = {}
    _translators = {}
    _hparamsets = {}

    @staticmethod
    def dataset(name=None):
        def decorator(dataset_fn: Callable[[], Tuple[Dataset, Dataset, Dataset]]):
            nonlocal name
            if name is None:
                name = dataset_fn.__name__
            Registry._datasets[name] = dataset_fn
            return dataset_fn
        return decorator

    @staticmethod
    def get_dataset(name) -> Tuple[Dataset, Dataset, Dataset]:
        return Registry._datasets[name]()

    @staticmethod
    def translator(name=None):
        def decorator(translator_fn: Callable[[], Translator]):
            nonlocal name
            if name is None:
                name = translator_fn.__name__
            Registry._translators[name] = translator_fn
            return translator_fn
        return decorator

    @staticmethod
    def get_translator(name, **kwargs) -> Translator:
        return Registry._translators[name](**kwargs)

    @staticmethod
    def hparamset(name=None):
        def decorator(hparamset_fn: Callable[[], HyperParamSet]):
            nonlocal name
            if name is None:
                name = hparamset_fn.__name__
            Registry._hparamsets[name] = hparamset_fn
            return hparamset_fn
        return decorator

    @staticmethod
    def get_hparamset(name) -> HyperParamSet:
        return Registry._hparamsets[name]()



import logging
from typing import Optional, List, Dict, Union, Any, Sequence
from trialbot.utils.move_to_device import move_to_device
import torch.nn
from ..trial_bot import TrialBot
from trialbot.data.iterators import RandomIterator
from ..updater import Updater

class TrainingUpdater(Updater):
    def __init__(self, dataset, translator, models, iterators, optims, device=-1, dry_run: bool = False, grad_clip_value: float = 0.):
        super().__init__(models, iterators, optims, device)
        self._dry_run = dry_run
        self.dataset = dataset
        self.translator = translator
        self.grad_clip_value = grad_clip_value

    def update_epoch(self):
        model, optim, iterator = self._models[0], self._optims[0], self._iterators[0]
        if iterator.is_end_of_epoch:
            self.stop_epoch()

        device = self._device
        model.train()
        optim.zero_grad()
        indices: Sequence[int] = next(iterator)
        tensor_list = [self.translator.to_tensor(self.dataset[index]) for index in indices]
        try:
            batch = self.translator.batch_tensor(tensor_list)
        except Exception as e:
            logging.getLogger(self.__class__.__name__).warning(
                f'skipping the preprocessing of the batch which raises an exception ... {str(e)}'
            )
            return None

        if device >= 0:
            batch = move_to_device(batch, device)

        output = model(**batch)
        if not self._dry_run:
            loss = output['loss']
            loss.backward()

            if self.grad_clip_value > 0:
                torch.nn.utils.clip_grad_value_(model.parameters(), self.grad_clip_value)
            optim.step()
        return output

    @classmethod
    def from_bot(cls, bot: TrialBot) -> 'TrainingUpdater':
        args, p, model = bot.args, bot.hparams, bot.model
        logger = bot.logger

        if hasattr(p, "OPTIM") and isinstance(p.OPTIM, str) and p.OPTIM.lower() == "sgd":
            logger.info(f"Using SGD optimzer with lr={p.SGD_LR}")
            optim = torch.optim.SGD(model.parameters(), p.SGD_LR, weight_decay=p.WEIGHT_DECAY)
        elif hasattr(p, "OPTIM") and isinstance(p.OPTIM, str) and p.OPTIM.lower() == "radam":
            from radam import RAdam
            optim = RAdam(model.parameters(), lr=p.ADAM_LR, weight_decay=p.WEIGHT_DECAY)
            logger.info(f"Using RAdam optimzer with lr={p.ADAM_LR}")
        else:
            logger.info(f"Using Adam optimzer with lr={p.ADAM_LR} and beta={str(p.ADAM_BETAS)}")
            optim = torch.optim.Adam(model.parameters(), p.ADAM_LR, p.ADAM_BETAS, weight_decay=p.WEIGHT_DECAY)

        device = args.device
        dry_run = args.dry_run
        repeat_iter = not args.debug
        shuffle_iter = not args.debug
        iterator = RandomIterator(len(bot.train_set), bot.hparams.batch_sz, shuffle=shuffle_iter, repeat=repeat_iter)
        if args.debug and args.skip:
            iterator.reset(args.skip)

        updater = cls(bot.train_set, bot.translator, model, iterator, optim, device, dry_run,
                      grad_clip_value=p.GRAD_CLIPPING)
        return updater





from typing import Optional, Tuple
import pickle
import logging
from datetime import datetime as dt
import redis

from trialbot.data.dataset import Dataset

class PickleDataset(Dataset):
    def __init__(self,
                 filename: str,
                 conn: Optional[Tuple[str, int, int]] = None,
                 prefix: str = None,
                 ignore_missing_data: bool = True,
                 ):
        super().__init__()
        self.filename = filename

        self.r = None
        if conn is not None:
            host, port, db = conn
            pool = redis.ConnectionPool(host=host, port=port, db=db)
            self.r = redis.Redis(connection_pool=pool)

        self.prefix = prefix
        self._data: Optional[list] = None
        self._ignore_missing = ignore_missing_data
        self.logger = logging.getLogger(self.__class__.__name__)

    def _read_data(self, force_reading_with_cache: bool = False):
        if self._data is None and (force_reading_with_cache or self.r is None):
            self.logger.info(f'Start reading file {self.filename} at {dt.now().strftime("%H:%M:%S")}')
            self._data = pickle.load(open(self.filename, 'rb'))
            if self.r is not None:
                for i, x in enumerate(self._data):
                    key = self.prefix + str(i)
                    self.r.set(key, pickle.dumps(x))
                    if i % 1000 == 0:
                        self.logger.info(f'Writing the {i}th example into cache')
            self.logger.info(f'End reading file {self.filename} at {dt.now().strftime("%H:%M:%S")}')

    def __len__(self):
        if self.r is not None:
            length = self.r.get(self.prefix + 'len')
            if length is not None:
                return int(length)

        self._read_data()
        return len(self._data)

    def get_example(self, i):
        if self.r is not None:
            example_ = self.r.get(self.prefix + str(i))
            if example_ is None and self._ignore_missing:
                self.logger.warning(f"Forced to ignore the missing data {i}")
                return None
            example = pickle.loads(example_)
            return example

        self._read_data()
        return self._data[i]










from trialbot.data.ns_vocabulary import NSVocabulary, PADDING_TOKEN, START_SYMBOL, END_SYMBOL, DEFAULT_OOV_TOKEN

# Since translators and fields depend on torch, they'd better not be imported here
# to make the data namespace clean and support pure python use.
#

from trialbot.data.dataset import Dataset, CompositionalDataset
from trialbot.data.datasets.file_dataset import FileDataset
from trialbot.data.datasets.tabular_dataset import TabSepFileDataset
from trialbot.data.datasets.jsonl_dataset import JsonLDataset
from trialbot.data.datasets.json_dataset import JsonDataset
from trialbot.data.datasets.index_dataset import IndexDataset
from trialbot.data.datasets.consecutive_lines_dataset import ConsecutiveLinesDataset
from trialbot.data.datasets.tsv_dataset import TSVDataset
from trialbot.data.datasets.pickle_dataset import PickleDataset
from trialbot.data.datasets.redis_dataset import RedisDataset
from trialbot.data.datasets.seq2dict_dataset import SequentialToDict

from trialbot.data.iterator import Iterator
from trialbot.data.iterators.random_iterator import RandomIterator
from trialbot.data.iterators.cluster_iterator import ClusterIterator
from trialbot.data.iterators.bucket_iterator import BucketIterator





# Load files for specific formats

import json
import bz2
import gzip
import os


def open_json(infile):
    if isinstance(infile, str):
        infile = open_file(infile)

    for json_str in infile:
        try:
            yield json.loads(json_str)

        except json.JSONDecodeError as e:
            continue


def open_file(filename: str):
    # assuming the data resides in a text file, if you want other file types
    # you'd better reimplement another reading utility.
    assert isinstance(filename, str)
    if filename.endswith('.bz2'):
        f = bz2.open(filename, 'rt')    # by default, bz2 and gzip use binary mode (either "r" == "rb"),
    elif filename.endswith('.gz'):
        f = gzip.open(filename, 'rt')
    else:
        f = open(filename)

    return f


def open_dir(dirname: str):
    for file in os.listdir(dirname):
        for line in open_file(file):
            yield line


def open_dir_or_file(objname: str):
    assert objname
    if os.path.isdir(objname):
        return open_dir(objname)
    else:
        return open_file(objname)




#
# Some code snippets are copied from chainer, whose license is therefore adhered below.
#
# Copyright (c) 2015 Preferred Infrastructure, Inc.
# Copyright (c) 2015 Preferred Networks, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import six
try:
    import numpy
    is_ndarray = lambda x: isinstance(x, numpy.ndarray)
except ImportError:
    import logging
    logging.warning('numpy not installed; dataset indexing with a numpy ndarray will be unavailable')
    is_ndarray = lambda x: False

class Dataset:

    """The basic implementation of the dataset abstraction.

    By default, a dataset is just a holder or connector to existing data files,
    which is required to provide two functions for the interface:

    1. the `__len__` operator which will tell the total dataset size
    2. the `__getitem__` operator which supports indexing and slicing
       over the raw examples.

    The default implementation has supported slicing over the datasets,
    which only requires subclass to implement data indexing only.
    """

    def __getitem__(self, index):
        """Returns an example or a sequence of examples.

        It implements the standard Python indexing and one-dimensional integer
        array indexing. It uses the :meth:`get_example` method by default, but
        it may be overridden by the implementation to, for example, improve the
        slicing performance.

        Args:
            index (int, slice, list or numpy.ndarray): An index of an example
                or indexes of examples.

        Returns:
            If index is int, returns an example created by `get_example`.
            If index is either slice or one-dimensional list or numpy.ndarray,
            returns a list of examples created by `get_example`.

        """
        if isinstance(index, slice):
            current, stop, step = index.indices(len(self))
            return [self.get_example(i) for i in
                    six.moves.range(current, stop, step)]
        elif isinstance(index, list) or is_ndarray(index):
            return [self.get_example(i) for i in index]
        else:
            return self.get_example(index)

    def __len__(self):
        raise NotImplementedError

    def get_example(self, i: int):
        """Returns the i-th example.
        However, we do not require an example be of any specific type.
        As long as the interpreter of the dataset could recognize it, that's enough.
        It's better for a subclass of dataset to be annotated with a proper type,
        for the sake of IDE intelligence over the accompanied interpreter,
        but this is not a must.

        Implementations should override it. It should raise :class:`IndexError`
        if the index is invalid.

        Args:
            i (int): The index of the example.

        Returns:
            The i-th example.

        """
        raise NotImplementedError

class CompositionalDataset(Dataset):
    def __init__(self, dataset: Dataset):
        self.dataset = dataset




from ..dataset import CompositionalDataset, Dataset
from typing import Optional, Tuple
import redis
import logging
import pickle

class RedisDataset(CompositionalDataset):
    def __init__(self,
                 dataset: Dataset,
                 conn: Optional[Tuple[str, int, int]] = None,
                 prefix: str = "",
                 expire_sec: int = 0,
                 ):
        super().__init__(dataset)
        self.logger = logging.getLogger(self.__class__.__name__)

        self.r = None
        if conn is not None:
            host, port, db = conn
            try:
                pool = redis.ConnectionPool(host=host, port=port, db=db)
                self.r = redis.Redis(connection_pool=pool)
                self.r.ping()
            except:
                self.logger.warning('specified redis server is not available, the redis dataset is used passthrough')
                self.r = None

        self.prefix = prefix if prefix.endswith('_') else prefix + '_'
        self.dataset = dataset
        self.expire_sec = expire_sec

    def _set(self, k, v):
        if self.r is None:
            return

        arr = pickle.dumps(v)
        self.r.set(k, arr, ex=self.expire_sec if self.expire_sec > 0 else None)

    def _read(self, k):
        if self.r is None:
            return None

        v = self.r.get(k)
        if v is not None:
            v = pickle.loads(v)
        return v

    def get_example(self, i: int):
        obj = self._read(self.prefix + str(i))
        if obj is None:
            obj = self.dataset.get_example(i)

        self._set(self.prefix + str(i), obj)
        return obj

    def __len__(self):
        l = self._read(self.prefix + 'len')
        if l is None:
            l = len(self.dataset)
        self._set(self.prefix + 'len', l)
        return l




from ..dataset import Dataset, CompositionalDataset
from typing import List

class SequentialToDict(CompositionalDataset):
    def __init__(self, dataset: Dataset, keys: List[str]):
        super().__init__(dataset)
        self.keys = keys

    def __len__(self):
        return len(self.dataset)

    def get_example(self, i: int):
        iterable_data = self.dataset.get_example(i)
        return dict(zip(self.keys, iterable_data))


from typing import Mapping, Set, Optional, List, Any
from ..translator import FieldAwareTranslator, NullableTensor


class _DictKeyRefWrapper(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ref_set = set()

    def get(self, k):
        if k not in self:
            return None
        return self[k]

    def __getitem__(self, item):
        self.ref_set.add(item)
        return super().__getitem__(item)


class KnownFieldTranslator(FieldAwareTranslator):
    """
    A Translator aware of fields,
    not only translates known keys of an instance into the tensors as the specified fields,
    but collects other unknown field of an instance and arrange them into an array.
    """
    def __init__(self, accessed_field_set=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._accessed_field_set: Optional[Set] = accessed_field_set

    def to_tensor(self, example) -> Mapping[str, NullableTensor]:
        output = {}
        if self._accessed_field_set is None:
            example = _DictKeyRefWrapper(example)
            output.update(super().to_tensor(example))
            self._accessed_field_set = set(example.ref_set)
        else:
            output.update(super().to_tensor(example))

        for field in filter(lambda k: k not in self._accessed_field_set, example.keys()):
            output[field] = example[field]

        return output

    def batch_tensor(self, tensors: List[Mapping[str, NullableTensor]]) -> Mapping[str, Any]:
        processed_fields = self._accessed_field_set or set()
        tensors = list(filter(lambda x: all(v is not None for v in x.values()), tensors))
        batch_dict = self.list_of_dict_to_dict_of_list(tensors)
        output = {}
        for field in self.fields:
            output.update(field.batch_tensor_by_key(batch_dict))
        for k, v in batch_dict.items():
            if k not in processed_fields and k not in output.keys():
                output[k] = v
        return output



from trialbot.data.datasets.file_dataset import FileDataset

class TSVDataset(FileDataset):
    def __init__(self, filename, sep="\t", has_title=True, lazy=True):
        super().__init__(filename, lazy)
        self._sep = sep
        self._has_title = has_title
        self._fields = None
        if (not self.lazy) and len(self._data) > 0:
            self._fields = self._data[0].split(self._sep)

    def get_example(self, i):
        example = super().get_example(i + 1)
        example_tuple = example.split(self._sep)
        if not self._has_title:
            return example_tuple

        if self._fields is None:
            self._fields = self._data[0].split(self._sep)

        example_dict = dict(zip(self._fields, example_tuple))
        return example_dict

    def __len__(self):
        length = super().__len__()
        if self._has_title:
            length -= 1
        return length



#
# This interface code is originated from chainer, whose license is therefore adhered below.
#
# Copyright (c) 2015 Preferred Infrastructure, Inc.
# Copyright (c) 2015 Preferred Networks, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
from typing import Sequence

class Iterator(object):
    """Base class of all dataset iterators.

    Iterator iterates over the dataset, yielding a sequence of indices at each iteration.
    The indices sequence usually indicates a list of examples of a minibatch.
    Each implementation should implement an iterator protocol (e.g., the :meth:`__next__` method).

    Each implementation should provide the following attributes (not needed to be writable).

    - ``epoch``: Number of completed sweeps over the dataset.
    - ``is_end_of_epoch``: ``True`` if the batch indices returned is the end of the epoch.

    """
    def __init__(self):
        self.is_end_of_epoch = False
        self.epoch = 0

    def __iter__(self):
        """Returns self."""
        return self

    def __next__(self) -> Sequence[int]:
        """Returns the next batch.

        This is a part of the iterator protocol of Python. It may raise the
        :class:`StopIteration` exception when it stops the iteration.

        """
        raise NotImplementedError

    def next(self):
        """Python2 alternative of ``__next__``.
        It calls :meth:`__next__` by default.
        """
        return self.__next__()

    def finalize(self):
        """Finalizes the iterator and possibly releases the resources.

        This method does nothing by default. Implementation may override it to
        better handle the internal resources.

        """
        pass

    def __enter__(self):
        """With statement context manager method

        This method does nothing by default. Implementation may override it to
        better handle the internal resources by with statement.

        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """With statement context manager method

        This method does nothing by default. Implementation may override it to
        better handle the internal resources by with statement.

        """
        return None





from typing import List, Mapping, Union, Callable, Any, Sequence, Optional

import random

from trialbot.data.iterator import Iterator

import logging

class BucketIterator(Iterator):
    def __init__(self,
                 dataset,
                 batch_size,
                 key: Union[str, Callable[[Any], float]], # only one key used to rank
                 sort_noise: float = .1,
                 repeat=True,
                 drop_last: bool = False,
                 shuffle=True,
                 ):
        super().__init__()
        self.dataset = dataset
        self._dataset_lengths = None
        self.batch_size = batch_size

        self._key_func: Callable[[...], float] = key
        if isinstance(key, str):
            self._key_func = lambda x: len(x[key])

        self.sort_noise = sort_noise

        self.drop_last = drop_last

        self.repeat = repeat
        self._shuffle = shuffle
        self.logger = logging.getLogger(self.__class__.__name__)
        self._current_bucket = 0
        self.epoch = 0
        self._batches: Optional[List[List[int]]] = None

    def __next__(self) -> Sequence[int]:
        if self.epoch_size == 0 or self.batch_size <= 0:
            self.logger.warning("Given dataset is empty. Nothing to yield but only raising StopIteration")
            raise StopIteration

        if not self.repeat and self.epoch > 0:
            raise StopIteration

        if self._batches is None:
            self._batches = self._get_batches()

        indices = self._batches[self._current_bucket]

        self.is_end_of_epoch = False
        self._current_bucket += 1

        if self._current_bucket >= len(self._batches):
            self._current_bucket = 0
            self.epoch += 1
            # the current batch is the last of the epoch, and the is_end_of_epoch flag is set True
            # so the epoch training must not stop immediately but after processing the returned batch.
            self.is_end_of_epoch = True
            self._batches = None    # buckets will be reshuffled for the next turn

        return indices

    @staticmethod
    def add_uniform_noise_to_value(value: int, noise_param: float):
        if noise_param > 0:
            noise_value = value * noise_param
            noise = random.uniform(-noise_value, noise_value)
            return value + noise
        else:
            return value

    def _argsort_dataset(self) -> List[int]:
        if self._dataset_lengths is None:
            self._dataset_lengths = [self._key_func(x) for x in self.dataset]

        noisy_lengths = [self.add_uniform_noise_to_value(v, self.sort_noise) for v in self._dataset_lengths]
        indices = sorted(range(self.epoch_size), key=lambda i: noisy_lengths[i])
        return indices

    def _get_batches(self) -> List[List[int]]:
        indices = self._argsort_dataset()
        buckets = []
        # split into buckets and randomly shuffle the buckets (the groups)
        for i in range(0, len(indices), self.batch_size):
            group = indices[i:i+self.batch_size]
            if len(group) < self.batch_size and self.drop_last:
                continue
            buckets.append(group)

        if self._shuffle:
            random.shuffle(buckets)

        return buckets

    next = __next__

    @property
    def epoch_size(self):
        return len(self.dataset)





class HyperParamSet:
    def __str__(self):
        obj = {}
        for attr in dir(self):
            if attr.startswith('_'):
                continue

            val = getattr(self, attr)
            if any(isinstance(val, t) for t in (tuple, list, dict, int, str, float)):
                obj[attr] = str(val)

        return "\n".join("%s: %s" % (k, v) for k, v in obj.items())

    @staticmethod
    def common_settings(root_path=None):
        import os.path
        hparams = HyperParamSet()

        hparams.DEVICE = -1
        ROOT = '.' if root_path is None else root_path
        hparams.ROOT = ROOT
        hparams.SNAPSHOT_PATH = os.path.join(ROOT, 'snapshots')

        hparams.LOG_REPORT_INTERVAL = (1, 'iteration')
        hparams.TRAINING_LIMIT = 500  # in num of epochs
        hparams.SAVE_INTERVAL = (100, 'iteration')
        hparams.batch_sz = 128

        hparams.ADAM_LR = 1e-3
        hparams.ADAM_BETAS = (.9, .98)
        hparams.ADAM_EPS = 1e-9

        hparams.WEIGHT_DECAY = 0.
        hparams.GRAD_CLIPPING = 2

        hparams.SGD_LR = 1e-2
        hparams.DATA_PATH = os.path.join(ROOT, 'data')

        hparams.NS_VOCAB_KWARGS = dict()

        return hparams




from .file_dataset import FileDataset

class TabSepFileDataset(FileDataset):
    def get_example(self, i):
        line = super(TabSepFileDataset, self).get_example(i)
        parts = line.rstrip('\r\n').split('\t')
        return tuple(parts)





from trialbot.data.datasets.file_dataset import FileDataset
import json

class JsonLDataset(FileDataset):
    def get_example(self, i):
        line = super().get_example(i)
        return json.loads(line)


from typing import List, Mapping, Generator, Tuple, Optional
from .ns_vocabulary import NSVocabulary
import torch

NullableTensor = Optional[torch.Tensor]

class _TranslationBase:
    def __init__(self):
        self.vocab: Optional[NSVocabulary] = None

    def generate_namespace_tokens(self, example) -> Generator[Tuple[str, str], None, None]:
        raise NotImplementedError

    def index_with_vocab(self, vocab: NSVocabulary):
        self.vocab = vocab

    def to_tensor(self, example) -> Mapping[str, NullableTensor]:
        raise NotImplementedError




from trialbot.data.datasets.file_dataset import FileDataset

class ConsecutiveLinesDataset(FileDataset):
    def __init__(self, filename, lazy=True, num_lines=2):
        super(ConsecutiveLinesDataset, self).__init__(filename, lazy=lazy)
        self._num_lines = num_lines

    def __len__(self):
        return super().__len__() // 2

    def get_example(self, i):
        example = tuple(
            super().get_example(idx)
            for idx in range(i * self._num_lines, (i + 1) * self._num_lines)
        )
        return example



import argparse

from .trial_registry import Registry


def get_trial_bot_common_opt_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()

    parser.add_argument('--seed', type=int, help='manually set the seeds for torch')
    parser.add_argument('--device', type=int, default=-1, help="the gpu device number to override")
    parser.add_argument("--quiet", action="store_true", help="mute the log")
    parser.add_argument("--debug", action="store_true", help="print the debugging log")
    parser.add_argument('--memo', type=str, default="", help="used to remember some runtime configurations")
    parser.add_argument('--test', action="store_true", help='use testing mode')
    parser.add_argument('--dev', action="store_true", help='use testing mode on dev data')
    parser.add_argument('--hparamset', '-p', help="choose one of the available hyper-parameters",
                        choices=Registry._hparamsets.keys())
    parser.add_argument('--snapshot-dir', help="snapshot dir if continues")
    parser.add_argument('--dataset', choices=Registry._datasets.keys())
    parser.add_argument('--translator', choices=Registry._translators.keys())

    return parser


import json

from trialbot.data.datasets.file_dataset import FileDataset

class JsonDataset(FileDataset):
    def _read_data(self):
        if self._data is None:
            self._data = json.load(open(self.filename, 'r'))


"""
Source copied from AllenNLP.
"""
import torch
def has_tensor(obj) -> bool:
    """
    Given a possibly complex data structure,
    check if it has any torch.Tensors in it.
    """
    if isinstance(obj, torch.Tensor):
        return True
    elif isinstance(obj, dict):
        return any(has_tensor(value) for value in obj.values())
    elif isinstance(obj, (list, tuple)):
        return any(has_tensor(item) for item in obj)
    else:
        return False


def move_to_device(obj, cuda_device: int):
    """
    Given a structure (possibly) containing Tensors on the CPU,
    move all the Tensors to the specified GPU (or do nothing, if they should be on the CPU).
    """
    if cuda_device < 0 or not has_tensor(obj):
        return obj
    elif isinstance(obj, torch.Tensor):
        return obj.cuda(cuda_device)
    elif isinstance(obj, dict):
        return {key: move_to_device(value, cuda_device) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [move_to_device(item, cuda_device) for item in obj]
    elif isinstance(obj, tuple):
        return tuple([move_to_device(item, cuda_device) for item in obj])
    else:
        return obj

